/*******************************************************************************
    Copyright (c) 2013 NVidia Corporation

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to
    deal in the Software without restriction, including without limitation the
    rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
    sell copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*******************************************************************************/

/*
       uvm-debug.h

       This file contains the debug API for UVM
*/

#ifndef _UVM_DEBUG_H_
#define _UVM_DEBUG_H_

#ifdef __cplusplus
extern "C" {
#endif

#include "uvmtypes.h"

#define UVM_DEBUG_V1    0x00000001

/*******************************************************************************
    Debugging API types
*******************************************************************************/

/*******************************************************************************
* Counter scope: It can be one of the following:
    - Single GPU for a process (UvmCounterScopeProcessSingleGpu)
    - Aggregate of all GPUs for a process (UvmCounterScopeProcessAllGpu)
    - Single GPU system-wide (UvmCounterScopeGlobalSingleGpu)
    (UvmCounterScopeGlobalSingleGpu is not supported for CUDA 6.0)

    Note: The user must not assume that the counter values are equal to zero
    at the time of enabling counters.
    Difference between end state counter value and start state counter value 
    should be used to find out the correct value over a given period of time.
*******************************************************************************/

typedef enum 
{
    UvmCounterScopeProcessSingleGpu = 0,
    UvmCounterScopeProcessAllGpu = 1,
    UvmCounterScopeGlobalSingleGpu = 2
} UvmCounterScope;

// These numbers assigned to the counter name are used to index their value in
// the counter array.
//
typedef enum
{
    UvmCounterNameBytesXferHtD = 0,         // host to device
    UvmCounterNameBytesXferDtH = 1,         // device to host
    UvmCounterNameCpuPageFaultCount = 2,
#ifdef __windows__
    UvmCounterNameWddmBytesXferBtH = 3,         // backing store to host
    UvmCounterNameWddmBytesXferHtB = 4,         // host to backing store
    UvmCounterNameWddmBytesXferDtB = 5,         // eviction (device to backing store)
    UvmCounterNameWddmBytesXferBtD = 6,         // restoration (backing store to device)
#endif
    UVM_TOTAL_COUNTERS
} UvmCounterName;

/*******************************************************************************
    UVM counter config structure

     * scope: Please see the UvmCounterScope  enum (above), for details.
     * name: Name of the counter. Please check UvmCounterName for list.
     * gpuid: Identifies the GPU for which the counter will be enabled/disabled
              This parameter is ignored in AllGpu scopes.
     * state: A value of 0 will disable the counter, a value of 1 will enable 
              the counter.
*******************************************************************************/
typedef struct
{
    UvmCounterScope scope;//UVM_DEBUG_V1
    UvmCounterName name;  //UVM_DEBUG_V1
    UvmGpuUuid gpuid;     //UVM_DEBUG_V1
    unsigned state;       //UVM_DEBUG_V1
} UvmCounterConfig;

#define UVM_COUNTER_CONFIG_STATE_DISABLE_REQUESTED  0
#define UVM_COUNTER_CONFIG_STATE_ENABLE_REQUESTED   1

typedef UvmUPtr UvmDebugSession;

typedef unsigned UvmDebugEventQueue;

/*******************************************************************************
    UVM event notification types
    
    This type enumerates the different reasons we can have for the general
    notification events.
    
    NotificationShutdown: It will be added to the queue when the UVM driver is 
    shutdown
    
*******************************************************************************/
    
typedef enum
{
    UvmDebugEventNotificationShutdown = 0
} UvmDebugEventNotificationType;

/*******************************************************************************
    UVM event migration reason
    
    This type enumerates the different reasons we can have to trigger 
    a migration
    
    ReasonPageFault:    CPU accessed region and forced migrate to CPU
    
    ReasonKernelLaunch: GPU launch caused migration to GPU
    
    ReasonPaging:       Either system or GPU are low on resources. Please use
                        residency information or query region to determine if
                        it is a KMD paging operation or CPU

    ReasonSpeculative:  Speculative operations that may follow any of the
                        reasons above 
    
*******************************************************************************/
typedef enum
{
    UvmDebugMigrationReasonPageFault = 0,
    UvmDebugMigrationReasonKernelLaunch = 1,
    UvmDebugMigrationReasonPaging = 2,
    UvmDebugMigrationReasonSpeculative = 3
} UvmDebugMigrationReason;

/*******************************************************************************
    UVM event migration residency
    
    This type enumerates the different locations where memory can reside.
    
    ResidencyGpuExclusive:    Resides in CPU
    
    ResidencyCpuExclusive:    Resides in GPU

    ResidencyShared: Shared between GPU/CPU
    
*******************************************************************************/

typedef enum
{
    UvmDebugMemResidencyGpuExclusive = 0,
    UvmDebugMemResidencyCpuExclusive = 1,
    UvmDebugMemResidencyShared = 2
} UvmDebugMemResidency;

/*******************************************************************************
    UVM event access type
    
    This type enumerates the different type of accesses in memory
    
    TypeRead:    Read access
    
    TypeWrite:   Write access
    
*******************************************************************************/

typedef enum {
    UvmDebugAccessTypeRead = 0,
    UvmDebugAccessTypeWrite = 1
} UvmDebugAccessType;

/*******************************************************************************
    UVM event architecture specific structs
    
    The following structs define saved register information for several 
    architectures

*******************************************************************************/

typedef struct
{
    unsigned long long registerIP;    // PC of the thread that caused the fault.
    unsigned long long registerSP;    // Stack pointer of the thread that caused the fault.
    unsigned long long registerBP;    // Base pointer of the thread that caused the fault.
} UvmDebugExecCtxX86;

typedef struct
{
    unsigned registerPC;
    unsigned registerSP;
    unsigned registerLR;
    unsigned registerCSPR;
    unsigned registerSPSR;
} UvmDebugExecCtxARMv7;

/*******************************************************************************
    UVM memory region
    
    The following struct defines a memory region

*******************************************************************************/

typedef struct 
{
    UvmUPtr address;            // Base pointer to the region UVM_DEBUG_V1
    UvmSize size;               // Size of the region UVM_DEBUG_V1
    unsigned long long stream;  // Associated stream id UVM_DEBUG_V1
    UvmGpuUuid gpu;             // UVM_DEBUG_V1
} UvmDebugMemoryRegion;

/*******************************************************************************
    UVM event Scope
    
    The following type enumerates the possible scopes for the events
    
    Migration: Memory migrations between GPU-CPU
    
    MemoryViolation: Memory violations, these are normally fatal.

    Notifications: General notifications

*******************************************************************************/

typedef enum
{
    UvmDebugEventScopeMigration = 0,
    UvmDebugEventScopeMemoryViolation = 1,
    UvmDebugEventScopeNotifications = 2
} UvmDebugEventScope;


/*******************************************************************************
    UVM event scope Migration
    
    The following type describes the Migration event scope
    
    reason: Reason for the migration
    
    address: Address to the VA that migrated
    
    byteCount: Number of bytes migrated
    
    oldResidency: Residency of the VA before migration.
    
    curResidencty: Residency after Migration
    
    stream: Stream owner at the time

    eventStartUsec: Timestamp at which the event that triggered the migration 
                happened. It is expressed in usecs since EPOCH in UTC format.
                Note: this is the time at which the event is queued, not when 
                the actual copy started.
    eventEndUsec: TimeStamp at which the migration was completed.
    
    NOTE: The following fields won't be populated unless the user specified that
          it needs them on UvmDebugEventEnable() The reason being is that
          getting these values has a performance impact. The units of these
          counters don't match eventStart/eventEnd
    
    copyStartPtimer: PTIMER timestamp at which the copy starts.
    
    copyEndPtimer:   PTIMER timestamp at which the copy completes.

*******************************************************************************/

typedef struct
{
    UvmDebugMigrationReason reason;     // UVM_DEBUG_V1
    UvmUPtr address;                    // UVM_DEBUG_V1
    UvmSize byteCount;                  // UVM_DEBUG_V1  
    UvmDebugMemResidency oldResidency;  // UVM_DEBUG_V1
    UvmDebugMemResidency curResidencty; // UVM_DEBUG_V1
    unsigned long long stream;
    /* These counters are in usecs since EPOCH */
    unsigned long long eventStartUsec;  // UVM_DEBUG_V1
    unsigned long long eventEndUsec;    // UVM_DEBUG_V1 
    /* These counters are in PTIMER units.*/
    unsigned long long copyStartPtimer; // UVM_DEBUG_V1
    unsigned long long copyEndPtimer;   // UVM_DEBUG_V1 
} UvmDebugEventMigration;

#define UVM_DEBUG_EVENT_MIGRATION_ENABLE_PTIMER

/*******************************************************************************
    UVM event scope Memory Violation
    
    The following type describes the Memory Violation scope structure

    If  the host  part  of an  application  tries to  access a managed memory
    location when the stream that owns it is active on a GPU, the access will
    result in a SIGBUS being sent to the host application. A debug client can
    sign up to track memory violation events by allocating an event queue and
    enabling the event scope  "UvmDebugEventScopeMemoryViolation" type on the
    queue.  The  UVM  kernel  driver  will place  fault records into the event
    queue when a Memory Violation occurs.

    type: Indicates the type of access that caused the violation
    
    address: Address to the page that triggered the violation
    
    timestamp: timestamp at which the violation happened.
               It is expressed in usecs since EPOCH in UTC format

    execCtxValid: Indicates if the following field execCtx is valid.
    
    execCtx: Platform specific structure with the contents of the register at 
          the point of dereferencing the access that caused the violation.
    
*******************************************************************************/

typedef struct
{
    UvmDebugAccessType  type;        // UVM_DEBUG_V1
    UvmUPtr address;                 // UVM_DEBUG_V1
    unsigned long long timestamp;    // UVM_DEBUG_V1
    unsigned threadId;               // UVM_DEBUG_V1
    unsigned processId;              // UVM_DEBUG_V1
    UvmBool execCtxValid;            // UVM_DEBUG_VXX
    union
    {
        UvmDebugExecCtxX86 x64;
        UvmDebugExecCtxARMv7 armv7;
    } execCtx;                       // UVM_DEBUG_VXX
} UvmDebugEventMemoryViolation;

/*******************************************************************************
    UVM event scope general notificaton
    
    The following type describes the general notification scope structure
    
    type: Indicates the type of notification
    
    timestamp: timestamp at which the notification was queued.
               It is expressed in usecs since EPOCH in UTC format

*******************************************************************************/

typedef struct
{
    UvmDebugEventNotificationType type; // UVM_DEBUG_V1
    unsigned long long timestamp;       // UVM_DEBUG_V1
} UvmDebugEventNotification;

/*******************************************************************************
    UVM event entry
    
    This structure defines the entries of event that will be stored
    
*******************************************************************************/
typedef struct
{
    UvmDebugEventScope scope;                   // UVM_DEBUG_V1
    union
    {
        UvmDebugEventNotification notification; // UVM_DEBUG_V1
        UvmDebugEventMigration migration;       // UVM_DEBUG_V1
        UvmDebugEventMemoryViolation violation; // UVM_DEBUG_V1
    } entry;
} UvmDebugEventEntry;

/*******************************************************************************
    UVM event flags

    These flags are passed as arguments to UvmDebugEventQueueCreate

    NON_BLOCKING:   When this flag is passed, the events stream will override
                    unread entries if the queue is full instead of halting the
                    log.

*******************************************************************************/

#define UVM_DEBUG_EVENT_QUEUE_FLAGS_DEFAULT       (0x00000000)
#define UVM_DEBUG_EVENT_QUEUE_FLAGS_NON_BLOCKING  (0x00000001)


/*******************************************************************************
    UVM stream info

    Structure used to get information about the running state of a stream

    - Running: Indicates if the stream is running.

*******************************************************************************/
typedef struct
{
    UvmBool running;  // UVM_DEBUG_V1
} UvmDebugStreamInfo;

/*******************************************************************************
    UvmDebugGetVersion
    
    Get the version number of the UVM debug library

    Arguments:
    
    
    Returns:

    Version Number
*******************************************************************************/
unsigned UvmDebugVersion(void);

/*******************************************************************************
    UvmDebugCreateSession
    
    Creates a handle for a debugging session.
    
    When the client initializes, it will pass in a process handle and get a
    session ID for itself. Subsequent calls to the UVM API will take in that 
    session ID.
    
    There are security requirements to this call. 
    One of the following must be true:
    1.  The session owner must be running as an elevated user
    2.  The session owner and target must belong to the same user and the 
        session owner is at least as privileged as the target.
    
    For CUDA 6.0 we can create at most 64 sessions per debugger process.

    Arguments:
    
        pid: (INPUT)
            Process id for which the debugging session will be created
        
        session: (OUTPUT)
            Handle to the debugging session associated to that pid.
    
    Error codes:
        UVM_INVALID_ARGUMENT: If the pid is invalid.
        UVM_NOT_PERMITTED: If it fails the security check.
        UVM_ERR_INSUFFICIENT_RESOURCES : If an attempt is made to allocate more
            than 64 sessions per process.

*******************************************************************************/
UVM_STATUS UvmDebugCreateSession (unsigned pid, UvmDebugSession * session);

/*******************************************************************************
    UvmDebugDestroySession
    
    Destroys a debugging session.
    
    Arguments:
    
        session: (INPUT)
            Handle to the debugging session associated to that pid.
    
    Error codes:
        UVM_INVALID_ARGUMENT: If the session is invalid.
        UVM_ERR_RETRY_LATER: If the debug session is in use by some other thread.
*******************************************************************************/
UVM_STATUS UvmDebugDestroySession(UvmDebugSession session);

/*******************************************************************************
    UvmDebugCountersEnable
    
    Enables the counters following the user specified configuration.
    
    The user must fill a list with the configuration of the counters it needs to
    either enable or disable. It can only enable one counter per line.
    
    The structure (UvmCounterConfig) has several fields:
     * scope: Please see the UvmCounterScope  enum (above), for details.
     * name: Name of the counter. Please check UvmCounterName for list.
     * gpuid: Identifies the GPU for which the counter will be enabled/disabled
              This parameter is ignored in AllGpu scopes.
     * state: A value of 0 will disable the counter, a value of 1 will enable 
              the counter.

     Note: All counters are refcounted, that means that a counter will only be
     disable when its refcount reached zero.
    
    Arguments:
    
        session: (INPUT)
            Handle to the debugging session.
        
        config: (INPUT)
            pointer to configuration list as per above.
        
        count: (INPUT)
            number of entries in the config list.
    
    Error codes:
        UVM_NOT_PERMITTED:  if it fails the security check 
        UVM_INVALID_ARGUMENT: If debugging session is invalid or if one of the 
            counter lines is invalid. If call returns this value, no action 
            specified by the config list will have taken effect.
        UVM_ERR_NOT_SUPPORTED: UvmCounterScopeGlobalSingleGpu is not supported 
            for CUDA 6.0
        UVM_ERR_RETRY_LATER: If the debug session is in use by some other thread.
*******************************************************************************/
UVM_STATUS UvmDebugCountersEnable(UvmDebugSession session, 
                                UvmCounterConfig * config,
                                unsigned count);

/*******************************************************************************
    UvmDebugGetCounterHandle

    Returns handle to a particular counter. This is an opaque handle that the 
    implementation uses in order to find your counter, later.
    This handle can be used in subsequent calls to UvmDebugGetCounterVal().

    Arguments:

        session: (INPUT)
            Handle to the debugging session.

        scope: (INPUT)
            Scope that will be mapped.

        counterName: (INPUT)
            Name of the counter in that scope.

        UvmGpuUuid: (INPUT)
            Gpuid of the scoped GPU. This parameter is ignored in AllGpu scopes.

        pCounterHandle: (OUTPUT)
            Handle to the counter address.

    Error codes:
        UVM_INVALID_ARGUMENT: If the specified scope/gpu pair or session id is 
            invalid
        UVM_ERR_NOT_SUPPORTED: UvmCounterScopeGlobalSingleGpu is not supported 
            for CUDA 6.0
        UVM_ERR_RETRY_LATER: If the debug session is in use by some other thread. 
*******************************************************************************/
UVM_STATUS UvmDebugGetCounterHandle(UvmDebugSession session,
                               UvmCounterScope scope,
                               UvmCounterName counterName,
                               UvmGpuUuid gpu,
                               UvmUPtr *pCounterHandle);

/*******************************************************************************
    UvmDebugGetCounterVal

    Returns the counter value specified by the counter name.

    Arguments:

        session: (INPUT)
            Handle to the debugging session.
        
        counterHandleArray: (INPUT)
            Array of counter handles
        
        handleCount: (INPUT)
            Number of handles in the pPCounterHandle array.

        counterValArray: (OUTPUT)
            Array of counter values corresponding to the handles.

    Error codes:
        UVM_INVALID_ARGUMENT: If any of the specified handles is invalid.
*******************************************************************************/
UVM_STATUS UvmDebugGetCounterVal(UvmDebugSession session,
                            UvmUPtr *counterHandleArray,
                            unsigned handleCount,
                            unsigned long long *counterValArray);
#ifdef __cplusplus
}
#endif

#endif // _UVM_DEBUG_H_

